@@ -38,6 +38,8 @@
 #include <pdal/SrsBounds.hpp>
 #include <pdal/PipelineManager.hpp>
 #include <pdal/private/SrsTransform.hpp>
+#include <pdal/private/OGRSpec.hpp>
+#include <pdal/Polygon.hpp>
 
 #include <arbiter/arbiter.hpp>
 #include <nlohmann/json.hpp>
@@ -62,11 +64,44 @@ namespace pdal
 
 using namespace stac;
 
+namespace
+{
+
+struct Args
+{
+    std::vector<RegEx> items;
+    std::vector<RegEx> catalogs;
+    std::vector<RegEx> collections;
+
+    NL::json properties;
+    NL::json readerArgs;
+    NL::json rawReaderArgs;
+
+    NL::json::array_t dates;
+    SrsBounds bounds;
+    OGRSpec ogr;
+    std::vector<std::string> assetNames;
+
+    SchemaUrls schemaUrls;
+
+    bool validateSchema;
+    int threads;
+};
+
+} // unnamed namespace
+
 struct StacReader::Private
 {
 public:
-    //processing
+    Private() : m_args(std::make_unique<Args>())
+    {}
+
     std::unique_ptr<Args> m_args;
+    StageFactory m_factory;
+    MergeFilter m_merge;
+    LogPtr m_log;
+
+    //processing
     std::unique_ptr<ThreadPool> m_pool;
     std::vector<Reader*> m_readerList;
     std::unique_ptr<connector::Connector> m_connector;
@@ -83,30 +118,21 @@ public:
     std::unique_ptr<Catalog::Filters> m_catFilters;
     std::unique_ptr<Collection::Filters> m_colFilters;
     std::unique_ptr<ItemCollection::Filters> m_icFilters;
-};
 
-struct StacReader::Args
-{
+    void addItem(stac::Item& item);
+    void handleNested(stac::Catalog& c);
+    void handleItem(NL::json stacJson, std::string filename);
+    void handleCatalog(NL::json stacJson, std::string catPath);
+    void handleCollection(NL::json stacJson, std::string catPath);
+    void handleItemCollection(NL::json stacJson, std::string icPath);
+    void initializeArgs();
+    void printErrors(stac::Catalog& c);
 
-    std::vector<RegEx> items;
-    std::vector<RegEx> catalogs;
-    std::vector<RegEx> collections;
+    void setLog(LogPtr log)
+    { m_log = log; }
 
-    NL::json properties;
-    NL::json readerArgs;
-    NL::json rawReaderArgs;
-
-    NL::json::array_t dates;
-    SrsBounds bounds;
-    std::vector<std::string> assetNames;
-
-    SchemaUrls schemaUrls;
-
-    bool validateSchema;
-    int threads;
-
-    NL::json m_query;
-    NL::json m_headers;
+    LogPtr log() const
+    { return m_log; }
 };
 
 static PluginInfo const stacinfo
@@ -116,12 +142,17 @@ static PluginInfo const stacinfo
     "http://pdal.io/stages/readers.stac.html",
 };
 
+namespace
+{
+
+} // unnamed namespace
+
+
 CREATE_STATIC_STAGE(StacReader, stacinfo)
 
 std::string StacReader::getName() const { return stacinfo.name; }
 
 StacReader::StacReader():
-    m_args(new StacReader::Args),
     m_p(new StacReader::Private)
 {};
 
@@ -129,57 +160,52 @@ StacReader::~StacReader(){};
 
 void StacReader::addArgs(ProgramArgs& args)
 {
-    m_args.reset(new StacReader::Args());
-
     //Filter options
     args.add("asset_names", "List of asset names to look for in data"
-        " consumption. Default: 'data'", m_args->assetNames, {"data"});
+        " consumption. Default: 'data'", m_p->m_args->assetNames, {"data"});
     args.add("date_ranges", "Date ranges to include in your search. Dates are"
         "formatted according to RFC 3339. Eg. dates'[[\"min1\",\"max1\"],...]'"
-        , m_args->dates);
+        , m_p->m_args->dates);
     args.add("bounds", "Bounding box to select stac items by. This will "
-        "propogate down through all readers being used.", m_args->bounds);
+        "propogate down through all readers being used.", m_p->m_args->bounds);
+    args.add("ogr", "OGR filter geometries to select stac items by.",
+        m_p->m_args->ogr);
     args.add("properties", "Map of STAC property names to regular expression "
         "values. ie. {\"pc:type\": \"(lidar|sonar)\"}. Selected items will "
-        "match all properties.", m_args->properties);
+        "match all properties.", m_p->m_args->properties);
     args.add("items", "List of Item ID regexes to select STAC items based on.",
-        m_args->items);
+        m_p->m_args->items);
     args.add("catalogs", "List of Catalog ID regexes to select STAC items "
-        "based on.", m_args->catalogs);
+        "based on.", m_p->m_args->catalogs);
     args.add("collections", "List of Collection ID regexes to select STAC "
-        "items based on.", m_args->collections);
+        "items based on.", m_p->m_args->collections);
     args.addSynonym("items", "item_ids");
     args.addSynonym("catalogs", "catalog_ids");
     args.addSynonym("collections", "collection_ids");
 
     //Reader options
     args.add("validate_schema", "Use JSON schema to validate your STAC objects."
-        " Default: false", m_args->validateSchema, false);
-    args.add("header", "Header fields to forward with HTTP requests",
-        m_args->m_headers);
-    args.add("query", "Query parameters to forward with HTTP requests",
-        m_args->m_query);
+        " Default: false", m_p->m_args->validateSchema, false);
     args.add("reader_args", "Map of reader arguments to their values to pass"
-        " through.", m_args->rawReaderArgs);
+        " through.", m_p->m_args->rawReaderArgs);
     args.add("requests", "Number of threads for fetching JSON files, Default: 8",
-        m_args->threads, 8);
+        m_p->m_args->threads, 8);
     args.addSynonym("requests", "threads");
 
     //Schema options
     args.add("catalog_schema_url", "URL of catalog schema you'd like to use for"
-        " JSON schema validation.", m_args->schemaUrls.catalog,
+        " JSON schema validation.", m_p->m_args->schemaUrls.catalog,
         "https://schemas.stacspec.org/v1.0.0/catalog-spec/json-schema/catalog.json");
     args.add("collection_schema_url", "URL of collection schema you'd like to use for"
-        " JSON schema validation.", m_args->schemaUrls.collection,
+        " JSON schema validation.", m_p->m_args->schemaUrls.collection,
         "https://schemas.stacspec.org/v1.0.0/collection-spec/json-schema/collection.json");
     args.add("feature_schema_url", "URL of feature schema you'd like to use for"
-        " JSON schema validation.", m_args->schemaUrls.item,
+        " JSON schema validation.", m_p->m_args->schemaUrls.item,
         "https://schemas.stacspec.org/v1.0.0/item-spec/json-schema/item.json");
-
 }
 
 
-void StacReader::addItem(Item& item)
+void StacReader::Private::addItem(Item& item)
 {
     std::string driver = item.driver();
 
@@ -201,48 +227,32 @@ void StacReader::addItem(Item& item)
     }
     reader->setOptions(item.options());
 
-    std::lock_guard<std::mutex> lock(m_p->m_mutex);
-    m_p->m_itemList.push_back(item.id());
+    std::lock_guard<std::mutex> lock(m_mutex);
+    m_itemList.push_back(item.id());
     reader->setLog(log());
     m_merge.setInput(*reader);
-    m_p->m_readerList.push_back(reader);
+    m_readerList.push_back(reader);
 }
 
-void StacReader::handleItem(NL::json stacJson, std::string itemPath)
+void StacReader::Private::handleItem(NL::json stacJson, std::string filename)
 {
-    Item item(stacJson, m_filename, *m_p->m_connector,
-        m_args->validateSchema, log());
+    Item item(stacJson, filename, *m_connector, m_args->validateSchema);
+
     log()->get(LogLevel::Debug) << "Found STAC Item " << item.id() << ".";
-    if (item.init(*m_p->m_itemFilters, m_args->rawReaderArgs, m_args->schemaUrls))
-    {
+    if (item.init(*m_itemFilters, m_args->rawReaderArgs, m_args->schemaUrls))
         addItem(item);
-    }
 }
 
-void StacReader::printErrors(Catalog& c)
-{
-    ErrorList errors = c.errors();
-    if (errors.size())
-    {
-        for (auto& p: errors)
-        {
-            log()->get(LogLevel::Error) << "Failure fetching '" << p.first
-                << "' with error '" << p.second << "'\n";
-        }
-        throw pdal_error("Errors found during initial processing of the Catalog.");
-    }
-}
-
-void StacReader::handleNested(Catalog& c)
+void StacReader::Private::handleNested(Catalog& c)
 {
     auto& subs = c.subs();
     for (auto& sub: subs)
     {
         //add sub col/cat ids to list for metadata bookkeeping
         if (sub->type() == GroupType::catalog)
-            m_p->m_catList.push_back(sub->id());
+            m_catList.push_back(sub->id());
         else if (sub->type() == GroupType::collection)
-            m_p->m_colList.push_back(sub->id());
+            m_colList.push_back(sub->id());
 
         //collect items from sub catalogs
         for (Item& item: sub->items())
@@ -251,17 +261,15 @@ void StacReader::handleNested(Catalog& c)
 }
 
 
-void StacReader::handleCatalog(NL::json stacJson, std::string catPath)
+void StacReader::Private::handleCatalog(NL::json stacJson, std::string catPath)
 {
-    Catalog c(stacJson, catPath, *m_p->m_connector, *m_p->m_pool,
-        m_args->validateSchema, log());
+    Catalog c(stacJson, catPath, *m_connector, *m_pool, m_args->validateSchema, log());
 
     // if init returns false, the collection has no items in itself or in
     // any sub-catalogs/collections.
-    if (c.init(*m_p->m_catFilters, m_args->rawReaderArgs,
-        m_args->schemaUrls, true))
+    if (c.init(*m_catFilters, m_args->rawReaderArgs, m_args->schemaUrls, true))
     {
-        m_p->m_catList.push_back(c.id());
+        m_catList.push_back(c.id());
         //iteracted
         handleNested(c);
         //collect items from root
@@ -272,17 +280,15 @@ void StacReader::handleCatalog(NL::json stacJson, std::string catPath)
     printErrors(c);
 }
 
-void StacReader::handleCollection(NL::json stacJson, std::string colPath)
+void StacReader::Private::handleCollection(NL::json stacJson, std::string colPath)
 {
-    Collection c(stacJson, colPath, *m_p->m_connector,
-        *m_p->m_pool, m_args->validateSchema, log());
+    Collection c(stacJson, colPath, *m_connector, *m_pool, m_args->validateSchema, log());
 
     // if init returns false, the collection has no items in itself or in
     // any sub-catalogs/collections.
-    if (c.init(*m_p->m_colFilters, m_args->rawReaderArgs,
-        m_args->schemaUrls, true))
+    if (c.init(*m_colFilters, m_args->rawReaderArgs, m_args->schemaUrls, true))
     {
-        m_p->m_colList.push_back(c.id());
+        m_colList.push_back(c.id());
         handleNested(c);
         //collect items from root
         for (Item& item: c.items())
@@ -292,18 +298,31 @@ void StacReader::handleCollection(NL::json stacJson, std::string colPath)
     printErrors(c);
 }
 
-void StacReader::handleItemCollection(NL::json stacJson, std::string icPath)
+void StacReader::Private::handleItemCollection(NL::json stacJson, std::string icPath)
 {
-    ItemCollection ic(stacJson, icPath, *m_p->m_connector,
-            m_args->validateSchema, log());
+    ItemCollection ic(stacJson, icPath, *m_connector, m_args->validateSchema);
 
-    if (ic.init(*m_p->m_icFilters, m_args->rawReaderArgs, m_args->schemaUrls))
+    if (ic.init(*m_icFilters, m_args->rawReaderArgs, m_args->schemaUrls))
     {
         for (auto& item: ic.items())
             addItem(item);
     }
 }
 
+void StacReader::Private::printErrors(Catalog& c)
+{
+    ErrorList errors = c.errors();
+    if (errors.size())
+    {
+        for (auto& p: errors)
+        {
+            log()->get(LogLevel::Error) << "Failure fetching '" << p.first
+                << "' with error '" << p.second << "'\n";
+        }
+        throw pdal_error("Errors found during initial processing of the Catalog.");
+    }
+}
+
 std::string listStr(std::string key, std::vector<RegEx> ids)
 {
     std::stringstream s;
@@ -315,12 +334,12 @@ std::string listStr(std::string key, std::vector<RegEx> ids)
     return s.str();
 }
 
-void StacReader::initializeArgs()
+void StacReader::Private::initializeArgs()
 {
-    m_p->m_itemFilters = std::make_unique<Item::Filters>();
-    m_p->m_catFilters = std::make_unique<Catalog::Filters>();
-    m_p->m_colFilters = std::make_unique<Catalog::Filters>();
-    m_p->m_icFilters = std::make_unique<ItemCollection::Filters>();
+    m_itemFilters = std::make_unique<Item::Filters>();
+    m_catFilters = std::make_unique<Catalog::Filters>();
+    m_colFilters = std::make_unique<Catalog::Filters>();
+    m_icFilters = std::make_unique<ItemCollection::Filters>();
 
     log()->get(LogLevel::Debug) << "Filters: " << std::endl;
     if (!m_args->items.empty())
@@ -328,7 +347,7 @@ void StacReader::initializeArgs()
         std::string itIdStr = listStr("Item Ids", m_args->items);
         log()->get(LogLevel::Debug) << itIdStr;
 
-        m_p->m_itemFilters->ids = m_args->items;
+        m_itemFilters->ids = m_args->items;
     }
 
     if (!m_args->catalogs.empty())
@@ -336,7 +355,7 @@ void StacReader::initializeArgs()
         std::string caIdStr = listStr("Catalog Ids", m_args->catalogs);
         log()->get(LogLevel::Debug) << caIdStr;
 
-        m_p->m_catFilters->ids = m_args->catalogs;
+        m_catFilters->ids = m_args->catalogs;
     }
 
     if (!m_args->collections.empty())
@@ -344,8 +363,8 @@ void StacReader::initializeArgs()
         std::string coIdStr = listStr("Collection Ids", m_args->collections);
         log()->get(LogLevel::Debug) << coIdStr;
 
-        m_p->m_itemFilters->collections = m_args->collections;
-        m_p->m_colFilters->ids = m_args->collections;
+        m_itemFilters->collections = m_args->collections;
+        m_colFilters->ids = m_args->collections;
     }
 
 
@@ -373,7 +392,7 @@ void StacReader::initializeArgs()
                 if (minTime > maxTime)
                     log()->get(LogLevel::Warning) << "Min date (" << minDate <<
                         ") is greater than Max date (" << maxDate << ").";
-                m_p->m_itemFilters->datePairs.push_back({ minTime, maxTime });
+                m_itemFilters->datePairs.push_back({ minTime, maxTime });
             }
             catch(NL::detail::type_error&)
             {
@@ -390,9 +409,10 @@ void StacReader::initializeArgs()
             throw pdal_error("Properties argument must be a valid JSON object.");
         log()->get(LogLevel::Debug) << "Property Pruning: " <<
             m_args->properties.dump() << std::endl;
-        m_p->m_itemFilters->properties = m_args->properties;
+        m_itemFilters->properties = m_args->properties;
     }
 
+    // There should be a check if both are specified
     if (!m_args->bounds.empty())
     {
 
@@ -400,8 +420,14 @@ void StacReader::initializeArgs()
             throw pdal_error("Supplied bounds are not valid.");
         log()->get(LogLevel::Debug) << "Bounds: " << m_args->bounds << std::endl;
 
-        m_p->m_itemFilters->bounds = m_args->bounds;
+        Polygon boundsPoly(m_args->bounds.to2d());
+        if (!m_args->bounds.spatialReference().empty())
+            boundsPoly.setSpatialReference(m_args->bounds.spatialReference());
+        m_itemFilters->bounds = boundsPoly;
     }
+    else if (m_args->ogr.size())
+        m_itemFilters->bounds = m_args->ogr.getPolygons()[0];
+
     if (!m_args->assetNames.empty())
     {
         std::stringstream s;
@@ -410,74 +436,45 @@ void StacReader::initializeArgs()
             s << name << ", ";
         s.seekp(-2, std::ios_base::end);
         s << "]" << std::endl;
-        auto it = m_p->m_itemFilters->assetNames.begin();
+        auto it = m_itemFilters->assetNames.begin();
         log()->get(LogLevel::Debug) << s.str();
-        m_p->m_itemFilters->assetNames.insert(it, m_args->assetNames.begin(),
-            m_args->assetNames.end());
+        m_itemFilters->assetNames.insert(it, m_args->assetNames.begin(), m_args->assetNames.end());
     }
 
     if (m_args->validateSchema)
         log()->get(LogLevel::Debug) <<
             "JSON Schema validation flag is set." << std::endl;
 
-    m_p->m_colFilters->itemFilters = m_p->m_itemFilters.get();
-
-    m_p->m_catFilters->itemFilters = m_p->m_itemFilters.get();
-    m_p->m_catFilters->colFilters = m_p->m_colFilters.get();
-
-    m_p->m_icFilters->itemFilters = m_p->m_itemFilters.get();
-
-}
-
-void StacReader::setConnectionForwards(StringMap& headers, StringMap& query)
-{
-    try
-    {
-        if (!m_args->m_headers.is_null())
-            headers = m_args->m_headers.get<StringMap>();
-    }
-    catch (const std::exception& err)
-    {
-        throwError(std::string("Error parsing 'headers': ") + err.what());
-    }
-
-    try
-    {
-        if (!m_args->m_query.is_null())
-            query = m_args->m_query.get<StringMap>();
-    }
-    catch (const std::exception& err)
-    {
-        throwError(std::string("Error parsing 'query': ") + err.what());
-    }
+    m_colFilters->itemFilters = m_itemFilters.get();
+    m_catFilters->itemFilters = m_itemFilters.get();
+    m_catFilters->colFilters = m_colFilters.get();
+    m_icFilters->itemFilters = m_itemFilters.get();
 }
 
 void StacReader::addDimensions(PointLayoutPtr layout)
 {
-    StageWrapper::addDimensions(m_merge, layout);
+    StageWrapper::addDimensions(m_p->m_merge, layout);
 }
 
 void StacReader::initialize()
 {
-    StringMap headers;
-    StringMap query;
-    setConnectionForwards(headers, query);
-    m_p->m_connector.reset(new connector::Connector(headers, query));
+    m_p->m_connector.reset(new connector::Connector(m_filespec));
 
-    m_p->m_pool.reset(new ThreadPool(m_args->threads));
-    initializeArgs();
+    m_p->m_pool.reset(new ThreadPool(m_p->m_args->threads));
+    m_p->setLog(log());
+    m_p->initializeArgs();
 
     NL::json stacJson = m_p->m_connector->getJson(m_filename);
 
     std::string stacType = StacUtils::jsonValue<std::string>(stacJson, "type");
     if (stacType == "Feature")
-        handleItem(stacJson, m_filename);
+        m_p->handleItem(stacJson, m_filename);
     else if (stacType == "Catalog")
-        handleCatalog(stacJson, m_filename);
+        m_p->handleCatalog(stacJson, m_filename);
     else if (stacType == "Collection")
-        handleCollection(stacJson, m_filename);
+        m_p->handleCollection(stacJson, m_filename);
     else if (stacType == "FeatureCollection")
-        handleItemCollection(stacJson, m_filename);
+        m_p->handleItemCollection(stacJson, m_filename);
     else
         throw pdal_error("Could not initialize STAC object of type " + stacType);
 
@@ -485,10 +482,10 @@ void StacReader::initialize()
     m_p->m_pool->stop();
 
     if (m_p->m_readerList.empty())
-        throw pdal_error("Reader list is empty after filtering.");
+        throwError("Reader list is empty after filtering.");
 
-    setInput(m_merge);
-    StageWrapper::initialize(m_merge);
+    setInput(m_p->m_merge);
+    StageWrapper::initialize(m_p->m_merge);
 }
 
 QuickInfo StacReader::inspect()
@@ -556,20 +553,20 @@ bool StacReader::processOne(PointRef& point)
 
 void StacReader::prepared(PointTableRef table)
 {
-    m_merge.prepare(table);
-    m_merge.setLog(log());
+    m_p->m_merge.prepare(table);
+    m_p->m_merge.setLog(log());
 }
 
 
 void StacReader::ready(PointTableRef table)
 {
-    StageWrapper::ready(m_merge, table);
+    StageWrapper::ready(m_p->m_merge, table);
 }
 
 
 PointViewSet StacReader::run(PointViewPtr view)
 {
-    return StageWrapper::run(m_merge, view);
+    return StageWrapper::run(m_p->m_merge, view);
 }
 
 void StacReader::done(PointTableRef)
